# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# ISC License <https://opensource.org/licenses/isc>

from ._common import BaseHandler


class XMLHandler(BaseHandler):
    """ The XML serialization handler.
    """

    name = "xml"
    packages = ("lxml",)
    options = {
        "root": None,
        "pretty": False,
        "xml_declaration": False,
        "encoding": "utf-8",
    }

    def on_lxml_dumps(self, lxml, config, dictionary, **kwargs):
        """ The `lxml <https://pypi.org/project/lxml/>`_ dumps method.

        :param module lxml: The ``lxml`` module
        :param class config: The instance's config class
        :param dict dictionary: The dictionary to serialize
        :param str root: The root tag of the xml document, defaults to the config
            instance's name, optional
        :param bool pretty: Pretty format the resulting xml document, defaults to
            False, optional
        :param bool xml_declaration: Add the xml declaration header to the resulting
            xml document, defaults to False, optional
        :param str encoding: The encoding to use for the resulting xml document,
            defaults to "utf-8", optional
        :returns: The XML serialization
        :rtype: str
        """

        # NOTE: lazy import of XMLParser because class requires lxml to exist on import
        from ..contrib.xml_parser import XMLParser

        root = kwargs.pop("root")
        if not isinstance(root, str):
            root = config.__name__

        return XMLParser.from_dict(dictionary, root=root).to_xml(
            pretty=kwargs.pop("pretty", False),
            xml_declaration=kwargs.pop("xml_declaration", False),
            encoding=kwargs.pop("encoding", "utf-8"),
        )

    def on_lxml_loads(self, lxml, config, content, **kwargs):
        """ The `lxml <https://pypi.org/project/lxml/>`_ loads method.

        :param module lxml: The ``lxml`` module
        :param class config: The loading config class
        :param str content: The content to deserialize
        :param str encoding: The encoding to read the given xml document as, defaults to
            "utf-8", optional
        :returns: The deserialized dictionary
        :rtype: dict
        """

        # NOTE: lazy import of XMLParser because class requires lxml to exist on import
        from ..contrib.xml_parser import XMLParser

        return XMLParser.from_xml(
            content, encoding=kwargs.pop("encoding", "utf-8")
        ).to_dict()



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# ISC License <https://choosealicense.com/licenses/isc>

import re
import base64
import typing
import itertools
import collections
from enum import Enum
from functools import lru_cache

import attr

from .constants import CONFIG_KEY, REGEX_TYPE_NAME


class Types(Enum):
    """ An enum which keeps a record of top-level type grouping names.

    .. note:: Specifically implemented for ease of use with the ``TYPE_MAPPINGS`` and
        helps reduce the use of **magic** strings used throughout the ``is_x_type``
        methods.
    """

    NULL = "null"
    BOOL = "bool"
    BYTES = "bytes"
    INTEGER = "integer"
    NUMBER = "number"
    STRING = "string"
    ARRAY = "array"
    OBJECT = "object"
    ENUM = "enum"
    UNION = "union"


COMPILED_PATTERN_TYPE = type(re.compile(""))
TYPE_MAPPINGS = {
    "builtins": {
        Types.NULL: (type(None),),
        Types.BOOL: (bool,),
        Types.BYTES: (bytes, bytearray),
        Types.INTEGER: (int,),
        Types.NUMBER: (float,),
        Types.STRING: (str,),
        Types.ARRAY: (list, tuple, set, frozenset),
        Types.OBJECT: (dict,),
        Types.ENUM: (Enum,),
    },
    "typing": {
        Types.STRING: (typing.AnyStr,),
        Types.ARRAY: (typing.List, typing.Tuple, typing.Set, typing.FrozenSet),
        Types.OBJECT: (typing.Dict, typing.ChainMap, typing.NamedTuple),
        Types.UNION: (typing.Union,),
    },
    "collections": {
        Types.STRING: (collections.UserString,),
        Types.ARRAY: (collections.UserList,),
        Types.OBJECT: (
            collections.ChainMap,
            collections.Counter,
            collections.OrderedDict,
            collections.UserDict,
        ),
    },
}


def _get_types(type_):
    """ Gathers all types within the ``TYPE_MAPPINGS`` for a specific ``Types`` value.

    :param Types type_: The type to retrieve
    :return: All types within the ``TYPE_MAPPINGS``
    :rtype: list
    """

    return list(
        itertools.chain.from_iterable(
            map(lambda x: TYPE_MAPPINGS[x].get(type_, []), TYPE_MAPPINGS)
        )
    )


def encode_bytes(bytes_):
    return base64.encodebytes(bytes_).decode("utf-8")


def decode_bytes(string):
    if is_string_type(type(string)):
        string = bytes(string, "utf-8")
    return base64.decodebytes(string)


def is_config_var(var):
    return (
        isinstance(var, (attr._make.Attribute, attr._make._CountingAttr))
        and hasattr(var, "metadata")
        and CONFIG_KEY in var.metadata
    )


def is_config_type(type_):
    return (
        isinstance(type_, type)
        and hasattr(type_, "__attrs_attrs__")
        and hasattr(type_, CONFIG_KEY)
    )


def is_config(config_instance):
    return isinstance(config_instance, object) and is_config_type(type(config_instance))


@lru_cache()
def is_compiled_pattern(compiled_pattern):
    return isinstance(compiled_pattern, COMPILED_PATTERN_TYPE)


@lru_cache()
def is_builtin_type(type_):
    # NOTE: supported builtin types
    return isinstance(type_, type) and getattr(type_, "__module__", None) == "builtins"


@lru_cache()
def is_enum_type(type_):
    return isinstance(type_, type) and issubclass(type_, tuple(_get_types(Types.ENUM)))


@lru_cache()
def is_typing_type(type_):
    return getattr(type_, "__module__", None) == "typing"


@lru_cache()
def is_collections_type(type_):
    return (
        isinstance(type_, type) and getattr(type_, "__module__", None) == "collections"
    )


@lru_cache()
def is_regex_type(type_):
    return (
        callable(type_)
        and type_.__name__ == REGEX_TYPE_NAME
        and hasattr(type_, "__supertype__")
        and is_compiled_pattern(type_.__supertype__)
    )


@lru_cache()
def is_union_type(type_):
    if is_typing_type(type_) and hasattr(type_, "__origin__"):
        # NOTE: union types can only be from typing module
        return type_.__origin__ in _get_types(Types.UNION)
    return False


@lru_cache()
def is_null_type(type_):
    return type_ in _get_types(Types.NULL)


@lru_cache()
def is_bool_type(type_):
    return type_ in _get_types(Types.BOOL)


@lru_cache()
def is_bytes_type(type_):
    return type_ in _get_types(Types.BYTES)


@lru_cache()
def is_string_type(type_):
    string_types = _get_types(Types.STRING)
    if is_typing_type(type_):
        return type_ in string_types or is_regex_type(type_)
    return type_ in string_types


@lru_cache()
def is_integer_type(type_):
    return type_ in _get_types(Types.INTEGER)


@lru_cache()
def is_number_type(type_):
    return type_ in _get_types(Types.NUMBER)


@lru_cache()
def is_array_type(type_):
    array_types = _get_types(Types.ARRAY)
    if is_typing_type(type_):
        return type_ in array_types or (
            hasattr(type_, "__origin__") and type_.__origin__ in array_types
        )
    return type_ in array_types


def is_object_type(type_):
    object_types = _get_types(Types.OBJECT)
    if is_typing_type(type_):
        return type_ in object_types or (
            hasattr(type_, "__origin__") and type_.__origin__ in object_types
        )
    return type_ in object_types


def typecast(type_, value):
    # NOTE: does not do any special validation of types before casting
    # will just raise errors on type casting failures
    if is_builtin_type(type_) or is_collections_type(type_) or is_enum_type(type_):
        # FIXME: move to Types enum and TYPE_MAPPING entry
        if is_bytes_type(type_):
            return decode_bytes(value)
        return type_(value)
    elif is_regex_type(type_):
        return typecast(str, value)
    elif is_typing_type(type_):
        base_type = type_.__extra__
        arg_types = type_.__args__

        if is_array_type(type_):
            if len(arg_types) == 1:
                item_type = arg_types[0]
                return base_type([typecast(item_type, item) for item in value])
            else:
                return base_type(value)
        elif is_object_type(type_):
            if len(arg_types) == 2:
                (key_type, item_type) = arg_types
                return base_type(
                    {
                        typecast(key_type, key): typecast(item_type, item)
                        for (key, item) in value.items()
                    }
                )
            else:
                return base_type(value)
        else:
            return base_type(value)
    else:
        return value



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# ISC License <https://opensource.org/licenses/isc>

import collections

import attr
from lxml import etree


@attr.s
class XMLParser(object):
    """ A custom XML parser which is reflective between xml and dictionaries.

    .. warning:: This parser **requires** ``type`` to be set on every value element.
        This is the only way the XML can be correctly parsed and understood by the
        loading config.

        So this parser is really effective at being reflective with itself. However, it
        might require hacking to make it work with xml documents that are not built from
        a config instance in the first place.
    """

    tree = attr.ib(type=etree.Element)

    @classmethod
    def _build_base(cls, element):
        """ Builds a base value from the given etree element.

        :param lxml.etree.Element element: The element to build from
        :return: The built value
        """

        return __builtins__[element.get("type")](element.text)

    @classmethod
    def _build_list(cls, list_element, dict_type=collections.OrderedDict):
        """ Builds a list entry from a list parent element.

        :param lxml.etree.Element list_element: The element to build from
        :param class dict_type: The dictionary class to build with, defaults to
            :class:`collections.OrderedDict`, optional
        :return: The built list
        :rtype: list
        """

        result = []
        for element in list_element:
            if "type" in element.attrib:
                result.append(cls._build_base(element))
            else:
                if all((_.tag == element.tag) for _ in element):
                    result.append(cls._build_list(element))
                else:
                    result.append(cls._build_dict(element, dict_type=dict_type))
        return result

    @classmethod
    def _build_dict(cls, root_element, dict_type=collections.OrderedDict):
        """ Builds a dictionary entry from a dictionary parent element.

        :param lxml.etree.Element root_element: The element to build from
        :param dict_type: The dictionary class to build with, defaults to
            :class:`collections.OrderedDict`, optional
        :return: The built dictionary
        :rtype: dict
        """

        result = dict_type()
        for element in root_element:
            if "type" in element.attrib:
                result[element.tag] = cls._build_base(element)
            else:
                if all((_.tag == element.tag) for _ in element):
                    result[element.tag] = cls._build_list(element, dict_type=dict_type)
                else:
                    result[element.tag] = cls._build_dict(element, dict_type=dict_type)
        return result

    @classmethod
    def _build_base_etree(cls, parent, key, value):
        """ Builds a element from a base dictionary entry.

        :param lxml.etree.Element parent: The parent element to add to
        :param str key: The key of the entry
        :param value: The value of th entry
        :return: The built element
        :rtype: lxml.etree.Element
        """

        element = etree.Element(key, type=type(value).__name__)
        element.text = str(value)
        return element

    @classmethod
    def _build_list_etree(cls, parent, items):
        """ Builds a element for a list of items.

        :param lxml.etree.Element parent: The parent element to add to
        :param list items: The list to build elements for
        :return: The built list element
        :rtype: lxml.etree.Element
        """

        for item in items:
            if isinstance(item, dict):
                dict_element = etree.Element(parent.tag)
                parent.append(cls._build_dict_etree(dict_element, item))
            elif isinstance(item, (list, tuple, set, frozenset)):
                parent.append(cls._build_list_etree(parent, item))
            else:
                parent.append(cls._build_base_etree(parent, parent.tag, item))
        return parent

    @classmethod
    def _build_dict_etree(cls, parent, dictionary):
        """ Builds a element for a dictionary.

        :param lxml.etree.Element: The parent element to add to
        :param dict dictionary: The dictionary to build from
        :return: The built dictionary element
        :rtype: lxml.etree.Element
        """

        for (key, value) in dictionary.items():
            if isinstance(value, dict):
                parent.append(cls._build_dict_etree(etree.Element(key), value))
            elif isinstance(value, (list, tuple, set, frozenset)):
                list_parent = etree.Element(key)
                parent.append(cls._build_list_etree(list_parent, value))
            else:
                parent.append(cls._build_base_etree(parent, key, value))
        return parent

    @classmethod
    def from_dict(cls, dictionary, root="root"):
        """ Create an instance of ``XMLParser`` from a given dictionary.

        :param dict dictionary: The dictionary to build from
        :param root: The name of the root element to use, defaults to "root", optional
        :return: The new ``XMLParser`` instance
        :rtype: XMLParser
        """

        return cls(cls._build_dict_etree(etree.Element(root), dictionary))

    @classmethod
    def from_xml(cls, content, encoding="utf-8"):
        """ Create an instance of ``XMLParser`` from some xml content.

        :param str content: The xml content to build from
        :param str encoding: The encoding to use for reading the xml content
        :return: The new ``XMLParser`` instance
        :rtype: XMLParser
        """

        parser = etree.XMLParser(encoding=encoding)
        return cls(etree.fromstring(content.encode(encoding), parser=parser))

    def to_dict(self, dict_type=collections.OrderedDict):
        """ Get the dictionary representation of the current parser.

        :param class dict_type: The dictionary type to build
        :return: The resulting dictionary
        :rtype: dict
        """

        return self._build_dict(self.tree, dict_type=dict_type)

    def to_xml(self, pretty=False, xml_declaration=False, encoding="utf-8"):
        """ Get the xml string of the current parser.

        :param bool pretty: Pretty format the resulting xml string
        :param bool xml_declaration: Add xml declaration header to resulting xml content
        :param str encoding: The encoding to use for the resulting xml content
        :return: The xml string of the current parser
        :rtype: str
        """

        return etree.tostring(
            self.tree,
            pretty_print=pretty,
            xml_declaration=xml_declaration,
            encoding=encoding.upper(),
        ).decode(encoding)



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# ISC License <https://choosealicense.com/licenses/isc>

from . import contrib, handlers, __version__
from .constants import CONFIG_KEY
from ._file_config import var, config, to_dict, validate, from_dict, make_config
from .schema_builder import Regex, build_schema



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# ISC License <https://opensource.org/licenses/isc>

from .ini_parser import INIParser
from .xml_parser import XMLParser

__all__ = ["INIParser", "XMLParser"]



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# ISC License <https://choosealicense.com/licenses/isc>

from ._common import BaseHandler


class JSONHandler(BaseHandler):
    """ The JSON serialization handler.
    """

    name = "json"
    packages = ("rapidjson", "ujson", "json")
    options = {"indent": None, "sort_keys": False}

    def on_json_dumps(self, json, config, dictionary, **kwargs):
        """ The :mod:`json` dumps method.

        :param module json: The ``json`` module
        :param class config: The instance's config class
        :param dict dictionary: The dictionary instance to serialize
        :param int indent: The amount of spaces to use for indentation,
            defaults to 0, optional
        :returns: The json serialization of the given ``dictionary``
        :rtype: str
        """

        return json.dumps(dictionary, **kwargs)

    def on_json_loads(self, json, config, content, **kwargs):
        """ The :mod:`json` loads method.

        :param module json: The ``json`` module
        :param class config: The loading config class
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return json.loads(content)

    def on_ujson_dumps(self, ujson, config, dictionary, **kwargs):
        """ The `ujson <https://pypi.org/project/ujson/>`_ dumps method.

        :param module ujson: The ``ujson`` module
        :param class config: The instance's config class
        :param dict dictionary: The dictionary instance to serialize
        :param int indent: The amount of spaces to use for indentation,
            defaults to 0, optional
        :returns: The json serialization of the given ``dictionary``
        :rtype: str
        """

        if not kwargs.get("indent", None):
            kwargs["indent"] = 0
        return ujson.dumps(dictionary, **kwargs)

    def on_ujson_loads(self, ujson, config, content, **kwargs):
        """ The `ujson <https://pypi.org/project/ujson/>`_ loads method.

        :param module ujson: The ``ujson`` module
        :param class config: The loading config class
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return ujson.loads(content)

    def on_rapidjson_dumps(self, rapidjson, config, dictionary, **kwargs):
        """ The `rapidjson <https://pypi.org/project/python-rapidjson/>`_ dumps method.

        :param module rapidjson: The ``rapidjson`` module
        :param class config: The instance's config class
        :param dict dictionary: The dictionary instance to serialize
        :param int indent: The amount of spaces to use for indentation,
            defaults to 0, optional
        :returns: The json serialization of the given ``dictionary``
        :rtype: str
        """

        return rapidjson.dumps(dictionary, **kwargs)

    def on_rapidjson_loads(self, rapidjson, config, content, **kwargs):
        """ The `rapidjson <https://pypi.org/project/python-rapidjson/>`_ loads method.

        :param module rapidjson: The ``rapidjson`` module
        :param class config: The loading config class
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return rapidjson.loads(content)



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# ISC License <https://opensource.org/licenses/isc>

import warnings

from ._common import BaseHandler
from ..contrib.ini_parser import INIParser


class INIHandler(BaseHandler):
    """ The INI serialization handler.

    .. important:: INI files **cannot** support arrays of mappings.
        Using these in your config and trying to dump to ini will throw a
        :class:`ValueError` and will fail serialization.

        If you need to use a config with arrays of mappings use
        `toml <https://github.com/toml-lang/toml>`_ instead.
        This use case is one of the many things toml does better than traditional
        config files.
    """

    name = "ini"
    packages = ("configparser",)
    options = {"root": None, "delimiter": ":", "empty_sections": False}

    def on_configparser_dumps(self, configparser, config, dictionary, **kwargs):
        """ The :mod:`configparser` dumps method.

        :param module configparser: The ``configparser`` module
        :param class config: The instance's config class
        :param dict dictionary: The dictionary instance to serialize
        :param str root: The top-level section of the ini file,
            defaults to ``config.__name__``, optional
        :param str delimiter: The delimiter character used for representing nested
            dictionaries, defaults to ":", optional
        :return: The ini serialization of the given ``dictionary``
        :rtype: str
        """

        root_section = kwargs.pop("root")
        if not isinstance(root_section, str):
            root_section = config.__name__

        delimiter = kwargs.pop("delimiter", ":")
        if delimiter in root_section:
            warnings.warn(
                f"root section {root_section!r} contains delimiter character "
                f"{delimiter!r}, loading from the resulting content will likely fail"
            )

        try:
            return INIParser.from_dict(
                dictionary,
                root_section=root_section,
                delimiter=kwargs.pop("delimiter", ":"),
                empty_sections=kwargs.pop("empty_sections", False),
            ).to_ini()
        except ValueError:
            raise ValueError("INI cannot handle this config, try using toml instead")

    def on_configparser_loads(self, configparser, config, content, **kwargs):
        """ The :mod:`configparser` loads method.

        :param module configparser: The ``configparser`` module
        :param class config: The loading config class
        :param str content: The content to deserialize
        :return: The deserialized dictionary
        :rtype: dict
        """

        return INIParser.from_ini(content).to_dict(
            delimiter=kwargs.pop("delimiter", ":")
        )



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# ISC License <https://opensource.org/licenses/isc>

import io
import re
import collections
import configparser

DEFAULT_DELIMITER = ":"


class INIParser(configparser.ConfigParser):
    """ A custom INI (config) parser which obeys the Mozilla files specification.

    .. important:: This parser conforms to what Mozilla says configuration files and
        their encoded values and types should look like.
        You can find their specification `here <https://bit.ly/2DksT5u>`_

    .. warning:: This parser is very unstable across the extremely wide range of ways
        that INI files can be represented. Mainly this was just created to be reflective
        with it's own results rather than building a new :mod:`configparser`.

        So building a config instance that parses things like ``tox.ini`` might take a
        bit of hacking to work.
    """

    # characters that require the string to be quoted
    requires_quotes = (" ", "\t", "=")
    quoted_string_regex = re.compile(r"^(\'.*\')|(\".*\")$")

    @classmethod
    def _encode_var(cls, var):
        """ Encodes a variable to the appropriate string format for ini files.

        :param var: The variable to encode
        :return: The ini representation of the variable
        :rtype: str
        """

        if isinstance(var, str):
            if any(_ in var for _ in cls.requires_quotes):
                # NOTE: quoted strings should just use '"' according to the spec
                return '"' + var.replace('"', '\\"') + '"'
            return var
        else:
            return str(var)

    @classmethod
    def _decode_var(cls, string):
        """ Decodes a given string into the appropriate type in Python.

        :param str string: The string to decode
        :return: The decoded value
        """

        str_match = cls.quoted_string_regex.match(string)
        if str_match:
            return string.strip("'" if str_match.groups()[0] else '"')
        elif string.isdigit():
            return int(string)
        elif string.lower() in ("true", "false"):
            return string.lower() == "true"
        elif string.lstrip("-").isdigit():
            try:
                return int(string)
            except ValueError:
                # case where we mistake something like "--0" a a int
                return string
        elif "." in string.lstrip("-"):
            try:
                return float(string)
            except ValueError:
                # one off case where we mistake a single "." as a float
                return string
        else:
            return string

    @classmethod
    def _build_dict(
        cls, parser_dict, delimiter=DEFAULT_DELIMITER, dict_type=collections.OrderedDict
    ):
        """ Builds a dictionary of ``dict_type`` given the ``parser._sections`` dict.

        :param dict parser_dict: The ``parser._sections`` mapping
        :param str delimiter: The delimiter for nested dictionaries,
            defaults to ":", optional
        :param class dict_type: The dictionary type to use for building the dict,
            defaults to :class:`collections.OrderedDict`, optional
        :return: The resulting dictionary
        :rtype: dict
        """

        result = dict_type()
        for (key, value) in parser_dict.items():
            if isinstance(value, dict):
                nestings = key.split(delimiter)

                # build nested dictionaries if they don't exist (up to 2nd to last key)
                base_dict = result
                for nested_key in nestings[:-1]:
                    if nested_key not in base_dict:
                        base_dict[nested_key] = dict_type()
                    base_dict = base_dict[nested_key]

                base_dict[nestings[-1]] = cls._build_dict(
                    parser_dict.get(key), delimiter=delimiter, dict_type=dict_type
                )
            else:
                if "\n" in value:
                    result[key] = [
                        cls._decode_var(_) for _ in value.lstrip("\n").split("\n")
                    ]
                else:
                    result[key] = cls._decode_var(value)
        return result

    @classmethod
    def _build_parser(
        cls,
        dictionary,
        parser,
        section_name,
        delimiter=DEFAULT_DELIMITER,
        empty_sections=False,
    ):
        """ Populates a parser instance with the content of a dictionary.

        :param dict dictionary: The dictionary to use for populating the parser instance
        :param configparser.ConfigParser parser: The parser instance
        :param str section_name: The current section name to add the dictionary keys to
        :param str delimiter: The nested dictionary delimiter character,
            defaults to ":", optional
        :param bool empty_sections: Flag to allow the representation of empty sections
            to exist, defaults to False, optional
        :return: The populated parser
        :rtype: configparser.ConfigParser
        """

        for (key, value) in dictionary.items():
            if isinstance(value, dict):
                nested_section = delimiter.join([section_name, key])
                is_empty = all(isinstance(_, dict) for _ in value.values())
                if not is_empty or empty_sections:
                    parser.add_section(nested_section)
                cls._build_parser(value, parser, nested_section, delimiter=delimiter)
            elif isinstance(value, (list, tuple, set, frozenset)):
                if any(isinstance(_, dict) for _ in value):
                    raise ValueError(
                        f"INI files cannot support arrays with mappings, "
                        f"found in key {key!r}"
                    )
                parser.set(
                    section_name, key, "\n".join(cls._encode_var(_) for _ in value)
                )
            else:
                parser.set(section_name, key, cls._encode_var(value))
        return parser

    @classmethod
    def from_dict(
        cls,
        dictionary,
        root_section="root",
        delimiter=DEFAULT_DELIMITER,
        empty_sections=False,
    ):
        """ Create an instance of ``INIParser`` from a given dictionary.

        :param dict dictionary: The dictionary to create an instance from
        :param str root_section: The root key of the ini content, defaults to "root",
            optional
        :param str delimiter: The delimiter character to use for nested dictionaries,
            defaults to ":", optional
        :param bool empty_sections: Flag to allow representation of empty sections to
            exist, defaults to False, optional
        :return: The new ``INIParser`` instance
        :rtype: INIParser
        """

        parser = cls()
        parser.add_section(root_section)
        return cls._build_parser(
            dictionary,
            parser,
            root_section,
            delimiter=delimiter,
            empty_sections=empty_sections,
        )

    @classmethod
    def from_ini(cls, content):
        """ Create an instance of ``INIParser`` from some ini content string.

        :param str content: The ini content to create an instance from
        :return: The new ``INIParser`` instance
        :rtype: INIParser
        """

        parser = cls()
        parser.read_string(content)
        return parser

    def to_dict(self, delimiter=DEFAULT_DELIMITER, dict_type=collections.OrderedDict):
        """ Get the dictionary representation of the current parser.

        :param str delimiter: The delimiter used for nested dictionaries,
            defaults to ":", optional
        :param class dict_type: The dictionary type to use for building the dictionary
            reperesentation, defaults to collections.OrderedDict, optional
        :return: The dictionary representation of the parser instance
        :rtype: dict
        """

        root_key = self.sections()[0]
        return self._build_dict(
            self._sections, delimiter=delimiter, dict_type=dict_type
        ).get(root_key, {})

    def to_ini(self):
        """ Get the ini string of the current parser.

        :return: The ini string of the current parser
        :rtype: str
        """

        fake_io = io.StringIO()
        self.write(fake_io)
        return fake_io.getvalue()



# Copyright (c) 2018 Stephen Bunn (stephen@bunn.io)
# ISC License <https://choosealicense.com/licenses/isc>

import codecs
import pathlib
import configparser

import setuptools

BASE_DIR = pathlib.Path(__file__).parent

config = configparser.ConfigParser()
config.read(BASE_DIR.joinpath("setup.cfg").as_posix())

try:
    metadata = config["metadata"]
except KeyError:
    raise KeyError(f"cannot run setup.py if setup.cfg is missing [metadata] section")

setuptools.setup(
    package_dir={"": "src"},
    packages=setuptools.find_packages("src"),
    package_data={"": ["LICENSE*", "README*"]},
    version=metadata["version"],
)
